@@ -27,9 +27,7 @@
 #ifndef __G_DEBUG_H__
 #define __G_DEBUG_H__
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
+G_BEGIN_DECLS 
 
 typedef enum {
   G_DEBUG_FATAL_WARNINGS = 1 << 0
@@ -55,9 +53,6 @@ GLIB_VAR guint _g_debug_flags;
 
 void _g_debug_init (void) G_GNUC_INTERNAL;
 
-#ifdef __cplusplus
-}
-#endif /* __cplusplus */
-
+G_END_DECLS
 
 #endif /* __G_DEBUG_H__ */
