@@ -93,6 +93,14 @@ value_free (gpointer boxed)
   g_free (value);
 }
 
+static gpointer
+gdate_copy (gpointer boxed)
+{
+  const GDate *date = (const GDate*) boxed;
+
+  return g_date_new_julian (g_date_get_julian (date));
+}
+
 static gpointer
 gstring_copy (gpointer boxed)
 {
@@ -172,6 +180,18 @@ g_value_array_get_type (void)
   return type_id;
 }
 
+GType
+g_date_get_type (void)
+{
+  static GType type_id = 0;
+
+  if (!type_id)
+    type_id = g_boxed_type_register_static ("GDate",
+					    (GBoxedCopyFunc) gdate_copy,
+					    (GBoxedFreeFunc) g_date_free);
+  return type_id;
+}
+
 GType
 g_strv_get_type (void)
 {
